/*
 Part Serial of the Processing project - http://processing.org
 Copyright (c) 2004-05 Ben Fry & Casey Reas
 TwinTeethMC is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 TwinTeethMC, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with TwinTeethMC so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
import java.util.Map;
import jssc.*;

/**
 Serial class inspired in Serial but implementing a RTCS/CTS software protocol
 to avoid printer firmware buffer overflow. Protocol as following:
 
 The DCE sends a RTS char to the DTE. DTE see if has enough space in the buffer. If so, send
 a CTS char to DCE. Then DCE send a block of PTRL_BLOCK_SIZE data. The block size has to be 
 equal at both sides. Then DCE will start the secuence again. 
 
 If DCE does not receive the CTS char in a WAIT_CTS_TIMEOUT time,  then DCE will retry to send 
 the RTS a MAX_RTS_RETRIES
 number of times, after that it will throw an execption.
 
 **/
public class MySerial implements SerialPortEventListener {
  public SerialPort port;
  byte[] buffer = new byte[4096];
  int inBuffer = 0;
  int readOffset = 0;
  volatile byte cts_flag = 0;
  volatile byte rts_flag = 0;

  private static final int PTRL_TX_RTS_STATE    = 0;
  private static final int PTRL_WAIT_CTS_STATE  = 1;
  private static final int PTRL_TIMEOUT_STATE   = 2;
  private static final int PTRL_TX_BLOCK_STATE  = 3;

  private int WAIT_CTS_TIMEOUT = 500; // in miliseconds
  private int MAX_RTS_RETRIES = 5;

  private static final byte RTS_CHAR = 0x12;
  private static final byte CTS_CHAR = 0x14;

  private static final int PTRL_BLOCK_SIZE = 32;
  private static final int PTRL_MAX_BLOCKS = 8;

  String[] block = new String[PTRL_MAX_BLOCKS];

  /////// CONSTRUCTORS 

  public MySerial(String portName, int baudRate) {
    this(portName, baudRate, 'N', 8, 1);
  }

  public MySerial(String portName, int baudRate, char parity, int dataBits, float stopBits) {

    // setup parity
    if (parity == 'O') {
      parity = SerialPort.PARITY_ODD;
    } else if (parity == 'E') {
      parity = SerialPort.PARITY_EVEN;
    } else if (parity == 'M') {
      parity = SerialPort.PARITY_MARK;
    } else if (parity == 'S') {
      parity = SerialPort.PARITY_SPACE;
    } else {
      parity = SerialPort.PARITY_NONE;
    }
    // setup stop bits
    int stopBitsIdx = SerialPort.STOPBITS_1;
    if (stopBits == 1.5f) {
      stopBitsIdx = SerialPort.STOPBITS_1_5;
    } else if (stopBits == 2) {
      stopBitsIdx = SerialPort.STOPBITS_2;
    }
    port = new SerialPort(portName);
    try {
      // the native open() call is not using O_NONBLOCK, so this might block for certain operations (see write())
      port.openPort();
      port.setParams(baudRate, dataBits, stopBitsIdx, parity);
      // we could register more events here
      port.addEventListener(this, SerialPort.MASK_RXCHAR);
    } 
    catch (SerialPortException e) {
      // this used to be a RuntimeException before, so stick with it
      throw new RuntimeException("Error opening serial port " + e.getPortName() + ": " + e.getExceptionType());
    }
  }

  /////// METHODS 
  /**
   ** sendCommand ()
   **/
  public void sendCommand(String cmd) {
    if (!active()) 
      return;
    write(RTS_CHAR);
    rts_flag = 1;
    write(cmd+"\n");
  }

  /**
   ** sendBufferedCommand (String cmd, int timeOut, int numRetries)
   **/
  public void sendBufferedCommand(String cmd, int timeOut, int numRetries) {
    WAIT_CTS_TIMEOUT = timeOut;
    MAX_RTS_RETRIES = numRetries;
    sendBufferedCommand(cmd);
  }

  /**
   ** sendBufferedCommand (String cmd)
   **/
  public void sendBufferedCommand(String cmd) {

    if (!active()) 
      return;

    int state;
    int waitCTSTimeOut;
    int rts_retries = 0;
    String response;

    // Split the command in blocks of a max length

    int num_blocks = 0;
    int offset = 0;
    int len  = 0;

    do {
      block[num_blocks] = cmd.substring (offset, Math.min(cmd.length(), offset + PTRL_BLOCK_SIZE));
      len = block[num_blocks].length();
      if (len > 0) {
        offset += len;
      }
    } 
    while (len > 0 && num_blocks++ < PTRL_MAX_BLOCKS);

    if (num_blocks >= PTRL_MAX_BLOCKS) {
      throw new RuntimeException("Error: command too long at sendRasterCommand()\n");
    }

    if (num_blocks <= 0) {
      return;
    }

    // Send the blocks

    int block_inx = 0;
    rts_retries = 0;
    waitCTSTimeOut = millis();
    cts_flag = rts_flag = 0;

    state = PTRL_TX_RTS_STATE;

    while (true) {
      switch(state) {
      case 0: // PTRL_TX_RTS_STATE
        write(RTS_CHAR);
        waitCTSTimeOut = millis();
        rts_flag = 1;
        state = PTRL_WAIT_CTS_STATE;
        break;
      case 1: // PTRL_WAIT_CTS_STATE
        if (millis() - waitCTSTimeOut >= WAIT_CTS_TIMEOUT) {
          cts_flag = rts_flag = 0;
          state = PTRL_TIMEOUT_STATE;
        } else {
          if (cts_flag == 1) {
            state = PTRL_TX_BLOCK_STATE;
          }
        }
        break;
      case 2: // PTRL_TIMEOUT_STATE
        Thread.yield();
        if (rts_retries++ > MAX_RTS_RETRIES) {
          throw new RuntimeException("Error: no printer response sending RTS command\n");
        }
        state = PTRL_TX_RTS_STATE;
        break;
      case 3: // PTRL_TX_BLOCK_STATE
        if (block_inx >= (num_blocks -1)) {
          write(block[block_inx]+"\n");
          return;
        }
        write(block[block_inx++]);
        cts_flag = rts_flag = 0;
        state = PTRL_TX_RTS_STATE;
        break;
      }
    }
  }

  /**
   * Return true if this port is still active and hasn't run
   * into any trouble.
   */
  public boolean active() {
    return port.isOpened();
  }

  public int available() {
    return (inBuffer-readOffset);
  }

  public void clear() {
    synchronized (buffer) {
      inBuffer = 0;
      readOffset = 0;
    }
  }

  public Map<String, String> getProperties(String portName) {
    return SerialPortList.getPortProperties(portName);
  }

  public String[] list() {
    // returns list sorted alphabetically, thus cu.* comes before tty.*
    // this was different with RXTX
    return SerialPortList.getPortNames();
  }

  public int read() {
    if (inBuffer == readOffset) {
      return -1;
    }
    synchronized (buffer) {
      int ret = buffer[readOffset++] & 0xFF;
      if (inBuffer == readOffset) {
        inBuffer = 0;
        readOffset = 0;
      }
      return ret;
    }
  }

  public byte[] readBytesUntil(int inByte) {
    if (inBuffer == readOffset) {
      return null;
    }
    synchronized (buffer) {
      // look for needle in buffer
      int found = -1;
      for (int i=readOffset; i < inBuffer; i++) {
        if (buffer[i] == (byte)inByte) {
          found = i;
          break;
        }
      }
      if (found == -1) {
        return null;
      }
      int toCopy = found-readOffset+1;
      byte[] dest = new byte[toCopy];
      System.arraycopy(buffer, readOffset, dest, 0, toCopy);
      readOffset += toCopy;
      if (inBuffer == readOffset) {
        inBuffer = 0;
        readOffset = 0;
      }
      return dest;
    }
  }

  public String readStringUntil(int inByte) {
    byte temp[] = readBytesUntil(inByte);
    if (temp == null) {
      return null;
    } else {
      return new String(temp);
    }
  }

  // Check if CTS char in buffer, mark flag, filter it and copy rest of bytes to destination

  private int checkCTSAndcopy(byte[] src, int srcpos, byte[] dest, int destpos, int len) {
    int i = 0;
    int j = 0;

    while (i < len) {
      if (src[srcpos + i] == CTS_CHAR && rts_flag == 1) {
        cts_flag = 1;
      } else {
        dest[destpos + j] = src[srcpos + i];
        j++;
      }
      i++;
    }
    return j;
  }

  public void serialEvent(SerialPortEvent event) {
    if (event.getEventType() == SerialPortEvent.RXCHAR) {
      int toRead;
      int copied;
      try {
        while (0 < (toRead = port.getInputBufferBytesCount ())) {
          // this method can be called from the context of another thread
          synchronized (buffer) {
            // enlarge buffer if necessary
            if (buffer.length < inBuffer+toRead) {
              byte temp[] = new byte[buffer.length<<1];
              System.arraycopy(buffer, 0, temp, 0, inBuffer);
              buffer = temp;
            }
            // read an array of bytes and copy it into our buffer
            byte[] read = port.readBytes(toRead);
            //System.arraycopy(read, 0, buffer, inBuffer,read.length);
            copied = checkCTSAndcopy(read, 0, buffer, inBuffer, read.length);
            inBuffer += copied;
          }
        }
      } 
      catch (SerialPortException e) {
        throw new RuntimeException("Error reading from serial port " + e.getPortName() + ": " + e.getExceptionType());
      }
    }
  }

  public void stop() {
    try {
      port.closePort();
    } 
    catch (SerialPortException e) {
      // ignored
    }
    inBuffer = 0;
    readOffset = 0;
  }

  public void write(byte[] src) {
    try {
      // this might block if the serial device is not yet ready (esp. tty devices under OS X)
      port.writeBytes(src);
      // we used to call flush() here
    } 
    catch (SerialPortException e) {
      throw new RuntimeException("Error writing to serial port " + e.getPortName() + ": " + e.getExceptionType());
    }
  }
  public void write(int src) {
    try {
      port.writeInt(src);
    } 
    catch (SerialPortException e) {
      throw new RuntimeException("Error writing to serial port " + e.getPortName() + ": " + e.getExceptionType());
    }
  }
  public void write(String src) {
    try {
      port.writeString(src);
    } 
    catch (SerialPortException e) {
      throw new RuntimeException("Error writing to serial port " + e.getPortName() + ": " + e.getExceptionType());
    }
  }
}

